package de.hft.stuttgart.citydoctor2.gui;

import de.hft.stuttgart.citydoctor2.gui.tree.ZipEntryListCell;
import de.hft.stuttgart.citydoctor2.gui.tree.ZipEntryNode;
import de.hft.stuttgart.citydoctor2.utils.Localization;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipArchive;
import de.hft.stuttgart.citydoctor2.zip.CityGmlZipEntry;
import de.hft.stuttgart.citydoctor2.zip.ZipEntryErrorType;
import javafx.application.Platform;
import javafx.fxml.FXML;
import javafx.fxml.FXMLLoader;
import javafx.scene.Scene;
import javafx.scene.control.Button;
import javafx.scene.control.Label;
import javafx.scene.control.ListView;
import javafx.scene.control.ProgressBar;
import javafx.scene.control.TextField;
import javafx.scene.control.TitledPane;
import javafx.scene.control.Tooltip;
import javafx.scene.image.Image;
import javafx.scene.image.ImageView;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;
import javafx.scene.layout.VBox;
import javafx.stage.Modality;
import javafx.stage.Stage;
import javafx.stage.Window;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;

import java.io.IOException;
import java.io.InputStream;

public class ZipEntryManager {

    private static final Logger logger = LogManager.getLogger(ZipEntryManager.class);

    private Stage stage;
    private CityGmlZipArchive archive;

    @FXML
    private TitledPane entriesPane;
    @FXML
    private TitledPane metadataPane;

    @FXML
    private Label subpathLbl;
    @FXML
    private TextField subpathValueTxt;

    @FXML
    private Label erroneousLbl;
    @FXML
    private Label erroneousValue;

    @FXML
    private Label filesizeLbl;
    @FXML
    private Label filesizeValue;

    @FXML
    private Label validatedLbl;
    @FXML
    private Label validatedValue;

    @FXML
    private Label objectCountLbl;
    @FXML
    private Label objectCountValue;

    @FXML
    private Button cancelBtn;

    @FXML
    private Button loadBtn;
    @FXML
    private ImageView loadImageView;

    @FXML
    private Button decompressBtn;
    @FXML
    private ImageView decompressImageView;

    @FXML
    private Button decompressAllBtn;
    @FXML
    private ImageView decompressAllImageView;

    @FXML
    private Button checkAllBtn;
    @FXML
    private ImageView checkAllImageView;

    @FXML
    private Button showReportBtn;
    @FXML
    private ImageView showReportImageView;

    @FXML
    private ListView<ZipEntryNode> entryList;
    @FXML
    private ProgressBar progress;
    @FXML
    private ProgressBar subProgress;

    private CityDoctorController controller;
    private int currentlyLoadedEntry = -1;
    private CheckDialog dialog;

    private String unknownValueText = Localization.getText("ZipEntryManager.unknownValue");

    public ZipEntryManager(Window parent, CityDoctorController controller) throws IOException {
        FXMLLoader loader = new FXMLLoader(ZipEntryManager.class.getResource("ZipEntryManager.fxml"));
        loader.setController(this);
        this.controller = controller;
        archive = controller.getZipArchive();
        VBox box = loader.load();
        stage = new Stage();
        stage.getIcons().add(new Image(MainWindow.class.getResourceAsStream("icons/CityDoctor-Logo-rot_klein.jpg")));
        stage.setScene(new Scene(box));
        stage.initOwner(parent);
        stage.initModality(Modality.APPLICATION_MODAL);
        stage.setTitle(Localization.getText("ZipEntryManager.title"));
        stage.getScene().addEventFilter(KeyEvent.KEY_PRESSED, (KeyEvent event) -> {
            if (event.getCode() == KeyCode.ESCAPE) {
                stage.close();
            }
        });
        dialog = new CheckDialog(null, parent, controller);
        dialog.showZipArchiveOptions(true);
    }

    public void initialize() {
        subProgress.setVisible(false);
        entryList.getSelectionModel().selectedItemProperty().addListener((obs, oldI, newI) -> {
            if (newI != null) {
                showMetadata(newI.getEntry());
            }
        });
        entryList.setCellFactory(param -> new ZipEntryListCell());
        setupButtons();
        applyLocalization();
        populateZipEntryList();
    }

    private void setupButtons() {
        try {
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/decompress_zipEntry.png")) {
                Image img = new Image(inStream);
                decompressImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/view_model.png")) {
                Image img = new Image(inStream);
                loadImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/decompress_all_zipEntries.png")) {
                Image img = new Image(inStream);
                decompressAllImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/check_all_zipEntries.png")) {
                Image img = new Image(inStream);
                checkAllImageView.setImage(img);
            }
            try (InputStream inStream = MainWindow.class.getResourceAsStream("icons/error_stat32x32.png")) {
                Image img = new Image(inStream);
                showReportImageView.setImage(img);
            }
        } catch (IOException ignored) {
            //ignore exceptions
        }

        decompressBtn.setOnAction(e -> {
            disableTaskButtons();
            setEntryListLocked(true);
            Thread t = new Thread(() -> {
                CityGmlZipEntry entry = entryList.getSelectionModel().getSelectedItem().getEntry();
                logger.info("Decompressing entry \"{}\"", entry.getDisplayName());
                controller.decompressZipEntry(entry, progressValue ->
                        Platform.runLater(() -> progress.setProgress(progressValue))
                );
                Platform.runLater(() -> {
                    populateZipEntryList(entryList.getSelectionModel().getSelectedIndex());
                    progress.setProgress(0);
                    setEntryListLocked(false);
                });
            });
            t.setDaemon(true);
            t.start();

        });

        loadBtn.setOnAction(e -> {
            disableTaskButtons();
            setEntryListLocked(true);
            try {
                int selectedIndex = entryList.getSelectionModel().getSelectedIndex();
                if (selectedIndex == currentlyLoadedEntry) {
                    stage.hide();
                    return;
                }
                CityGmlZipEntry entry = entryList.getSelectionModel().getSelectedItem().getEntry();
                if (entry.getErrorType() != null) {
                    /**
                     * Button should be disabled when an erroneous entry gets selected.
                     * Silently fail in case an erroneous entry is somehow selected on Action.
                     */
                    logger.debug("IllegalState: loadBtn was pressed while an erroneous entry was selected");
                    populateZipEntryList(selectedIndex);
                    return;
                }
                controller.loadZipEntry(entry);
                currentlyLoadedEntry = selectedIndex;
                stage.hide();
            } finally {
                setEntryListLocked(false);
            }

        });

        decompressAllBtn.setOnAction(e -> {
            disableTaskButtons();
            setEntryListLocked(true);
            subProgress.setVisible(true);
            subProgress.setProgress(0);
            Thread t = new Thread(() -> {
                progress.setProgress(0);
                int count = 0;
                int total = entryList.getItems().size();
                for (ZipEntryNode node : entryList.getItems()) {
                    logger.info("Decompressing entry \"{}\"", node.getEntry().getDisplayName());
                    controller.decompressZipEntry(node.getEntry(), progressValue ->
                            Platform.runLater(() -> subProgress.setProgress(progressValue))
                    );
                    count++;
                    double progressValue = count * 1.0 / total;
                    Platform.runLater(() -> progress.setProgress(progressValue));
                }
                Platform.runLater(() -> {
                    subProgress.setVisible(false);
                    populateZipEntryList(entryList.getSelectionModel().getSelectedIndex());
                    progress.setProgress(0);
                    setEntryListLocked(false);
                });
            });
            t.setDaemon(true);
            t.start();
        });

        checkAllBtn.setOnAction(e -> {
            disableTaskButtons();
            dialog.show();
            refresh();
        });

        showReportBtn.setOnAction(e -> {
            disableTaskButtons();
            setEntryListLocked(true);

            try {
                controller.loadZipEntry(entryList.getSelectionModel().getSelectedItem().getEntry());
                if (currentlyLoadedEntry == -1) {
                    currentlyLoadedEntry = entryList.getSelectionModel().getSelectedIndex();
                }
                WriteReportDialog writeDialog = new WriteReportDialog(stage, controller, null);
                writeDialog.show();
            } catch (IOException ex) {
                throw new RuntimeException(ex);
            } finally {
                controller.loadZipEntry(entryList.getItems().get(currentlyLoadedEntry).getEntry());
                setEntryListLocked(false);
                refresh();
            }
        });
        showReportBtn.setDisable(true);

        cancelBtn.setOnAction(e -> stage.close());
    }

    private void applyLocalization() {
        entriesPane.setText(Localization.getText("ZipEntryManager.entryListTitle"));
        metadataPane.setText(Localization.getText("ZipEntryManager.metadata"));
        subpathLbl.setText(Localization.getText("ZipEntryManager.subpathLbl"));
        erroneousLbl.setText(Localization.getText("ZipEntryManager.erroneousLbl"));
        filesizeLbl.setText(Localization.getText("ZipEntryManager.filesizeLbl"));
        validatedLbl.setText(Localization.getText("ZipEntryManager.validatedLbl"));
        objectCountLbl.setText(Localization.getText("ZipEntryManager.objectCountLbl"));

        subpathValueTxt.setText(unknownValueText);
        erroneousValue.setText(unknownValueText);
        filesizeValue.setText(unknownValueText);
        validatedValue.setText(unknownValueText);
        objectCountValue.setText(unknownValueText);

        loadBtn.setTooltip(new Tooltip(Localization.getText("ZipEntryManager.loadBtn")));
        decompressBtn.setTooltip(new Tooltip(Localization.getText("ZipEntryManager.decompressBtn")));
        decompressAllBtn.setTooltip(new Tooltip(Localization.getText("ZipEntryManager.decompressAllBtn")));
        checkAllBtn.setTooltip(new Tooltip(Localization.getText("ZipEntryManager.checkAllBtn")));
        showReportBtn.setTooltip(new Tooltip(Localization.getText("ZipEntryManager.showReportBtn")));
        cancelBtn.setText(Localization.getText("ZipEntryManager.cancelBtn"));
    }

    private void populateZipEntryList() {
        populateZipEntryList(0);
    }

    private void populateZipEntryList(int selectionIndex) {
        entryList.getItems().clear();
        for (CityGmlZipEntry entry : archive.getEntries()) {
            entryList.getItems().add(new ZipEntryNode(entry));
        }
        entryList.getSelectionModel().clearSelection();
        entryList.getSelectionModel().select(selectionIndex);
        showMetadata(entryList.getSelectionModel().getSelectedItem().getEntry());
        decompressAllBtn.setDisable(
                entryList.getItems().stream().allMatch(zipEntryNode ->
                        (zipEntryNode.getEntry().isDecompressed()) || zipEntryNode.getEntry().getErrorType() != null));
    }

    private void showMetadata(CityGmlZipEntry entry) {
        subpathValueTxt.setText(entry.getFullFileName());
        if (entry.getFileSize() != -1L) {
            long fileSize = entry.getFileSize();
            long kb = 1024L;
            long mb = kb * kb;
            long gb = mb * kb;

            if (fileSize < kb) {
                filesizeValue.setText("< 1 KB");
            } else if (fileSize < mb) {
                filesizeValue.setText(String.format("%s KB", fileSize / kb));
            } else if (fileSize < gb) {
                filesizeValue.setText(String.format("%s MB", fileSize / mb));
            } else {
                double gigabytes = (double) fileSize / gb;
                filesizeValue.setText(String.format("%.2f GB", gigabytes));
            }
        } else {
            filesizeValue.setText(unknownValueText);
        }
        if (entry.getErrorType() != null) {
            erroneousValue.setText(getErrorText(entry.getErrorType()));
            validatedValue.setText(unknownValueText);
            objectCountValue.setText(unknownValueText);
            loadBtn.setDisable(true);
            decompressBtn.setDisable(true);
            showReportBtn.setDisable(true);
        } else {
            erroneousValue.setText(Localization.getText("ZipEntryManager.no"));

            loadBtn.setDisable(!entry.isDecompressed());
            decompressBtn.setDisable(entry.isDecompressed());
            if (entry.getModel() == null) {
                validatedValue.setText(unknownValueText);
                objectCountValue.setText(unknownValueText);
                return;
            }
            if (entry.getModel().isValidated()) {
                validatedValue.setText(Localization.getText("ZipEntryManager.yes"));
                showReportBtn.setDisable(false);
            } else {
                validatedValue.setText(Localization.getText("ZipEntryManager.no"));
                showReportBtn.setDisable(true);
            }
            objectCountValue.setText(String.valueOf(entry.getModel().getNumberOfFeatures()));
        }

    }

    private void disableTaskButtons() {
        decompressBtn.setDisable(true);
        loadBtn.setDisable(true);
        decompressAllBtn.setDisable(true);
    }

    public void refresh() {
        populateZipEntryList(entryList.getSelectionModel().getSelectedIndex());
    }

    private String getErrorText(ZipEntryErrorType error) {
        switch (error) {
            case EXCESSIVE_FILESIZE -> {
                return Localization.getText("ZipEntryManager.excessiveFileSize");
            }
            case INVALID_CITY_GML_FILE -> {
                return Localization.getText("ZipEntryManager.invalidCityGml");
            }
            default -> {
                return Localization.getText("ZipEntryManager.ioError");
            }
        }
    }

    private void setEntryListLocked(boolean locked) {
        entryList.setMouseTransparent(locked);
        entryList.setFocusTraversable(!locked);
    }

    public void show() {
        Platform.runLater(() -> {
            stage.showAndWait();
        });
        populateZipEntryList(entryList.getSelectionModel().getSelectedIndex());
    }
}
